# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the parametrization-related classes."""

from enum import Enum, unique

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.typing import Real


class ParamUV:
    """
    Parameter class containing 2 parameters: (u, v).

    Notes
    -----
    Likened to a 2D point in UV space Used as an argument in parametric
    surface evaluations. This matches the service implementation for the
    Geometry service.

    Parameters
    ----------
    u : Real
        u-parameter.
    v : Real
        v-parameter.
    """

    def __init__(self, u: Real, v: Real) -> None:
        """Initialize ``ParamUV`` class."""
        self._u = u
        self._v = v

    @property
    def u(self) -> Real:
        """u-parameter."""
        return self._u

    @property
    def v(self) -> Real:
        """v-parameter."""
        return self._v

    @check_input_types
    def __add__(self, other: "ParamUV") -> "ParamUV":
        """
        Add the u and v components of the other ParamUV to this ParamUV.

        Parameters
        ----------
        other : ParamUV
            The parameters to add these parameters.

        Returns
        -------
        ParamUV
            The sum of the parameters.
        """
        return ParamUV(self._u + other._u, self._v + other._v)

    @check_input_types
    def __sub__(self, other: "ParamUV") -> "ParamUV":
        """
        Subtract the u and v components of the other ParamUV from this ParamUV.

        Parameters
        ----------
        other : ParamUV
            The parameters to subtract from these parameters.

        Returns
        -------
        ParamUV
            The difference of the parameters.
        """
        return ParamUV(self._u - other._u, self._v - other._v)

    @check_input_types
    def __mul__(self, other: "ParamUV") -> "ParamUV":
        """
        Multiplies the u and v components of this ParamUV by the other ParamUV.

        Parameters
        ----------
        other : ParamUV
            The parameters to multiply by these parameters.

        Returns
        -------
        ParamUV
            The product of the parameters.
        """
        return ParamUV(self._u * other._u, self._v * other._v)

    @check_input_types
    def __truediv__(self, other: "ParamUV") -> "ParamUV":
        """
        Divides the u and v components of this ParamUV by the other ParamUV.

        Parameters
        ----------
        other : ParamUV
            The parameters to divide these parameters by.

        Returns
        -------
        ParamUV
            The quotient of the parameters.
        """
        return ParamUV(self._u / other._u, self._v / other._v)

    def __repr__(self) -> str:
        """Represent the ``ParamUV`` as a string."""
        return f"ParamUV(u={self.u}, v={self.v})"


class Interval:
    """
    Interval class that defines a range of values.

    Parameters
    ----------
    start : Real
        Start value of the interval.
    end : Real
        End value of the interval.
    """

    @check_input_types
    def __init__(self, start: Real, end: Real) -> None:
        """Initialize ``Interval`` class."""
        if end < start:
            raise ValueError("Start value must be less than end value")

        self._start = start
        self._end = end

    @property
    def start(self) -> Real:
        """Start value of the interval."""
        return self._start

    @property
    def end(self) -> Real:
        """End value of the interval."""
        return self._end

    def is_open(self) -> bool:
        """
        If the interval is open (-inf, inf).

        Returns
        -------
        bool
            True if both ends of the interval are negative and positive infinity respectively.
        """
        return np.isneginf(self.start) and np.isinf(self.end)

    def is_closed(self) -> bool:
        """
        If the interval is closed. Neither value is inf or -inf.

        Returns
        -------
        bool
            True if neither bound of the interval is infinite.
        """
        return self.start > np.NINF and self.end < np.inf

    def get_span(self) -> Real:
        """
        Return the quantity contained by the interval. Interval must be closed.

        Returns
        -------
        Real
            The difference between the end and start of the interval.
        """
        if not self.is_closed():
            raise ValueError("Interval must be closed to get the span.")

        return self.end - self.start

    def __repr__(self) -> str:
        """Represent the ``Interval`` as a string."""
        return f"Interval(start={self.start}, end={self.end})"


@unique
class ParamForm(Enum):
    """ParamForm enum class that defines the form of a Parameterization."""

    OPEN = 1
    CLOSED = 2
    PERIODIC = 3
    OTHER = 4


@unique
class ParamType(Enum):
    """ParamType enum class that defines the type of a Parameterization."""

    LINEAR = 1
    CIRCULAR = 2
    OTHER = 3


class Parameterization:
    """
    Parameterization class describes the parameters of a specific geometry.

    Parameters
    ----------
    form : ParamForm
        Form of the parameterization.
    type : ParamType
        Type of the parameterization.
    interval : Interval
        Interval of the parameterization.
    """

    @check_input_types
    def __init__(self, form: ParamForm, type: ParamType, interval: Interval) -> None:
        """Initialize ``Parameterization`` class."""
        self._form = form
        self._type = type
        self._interval = interval

    @property
    def form(self) -> ParamForm:
        """The form of the parameterization."""
        return self._form

    @property
    def type(self) -> ParamType:
        """The type of the parameterization."""
        return self._type

    @property
    def interval(self) -> Interval:
        """The interval of the parameterization."""
        return self._interval

    def __repr__(self) -> str:
        """Represent the ``Parameterization`` as a string."""
        return f"Parameterization(form={self.form}, type={self.type}, interval={self.interval})"



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides various measurement-related classes."""

from threading import Lock

from beartype import beartype as check_input_types
from beartype.typing import Optional, Union
from pint import Quantity, Unit

from ansys.geometry.core.misc.checks import check_is_float_int, check_pint_unit_compatibility
from ansys.geometry.core.misc.units import UNITS, PhysicalQuantity
from ansys.geometry.core.typing import Real


class SingletonMeta(type):
    """Provides a thread-safe implementation of a singleton design pattern."""

    # This class has been extracted from
    # https://refactoring.guru/design-patterns/singleton/python/example#example-1--main-py

    _instances = {}

    _lock: Lock = Lock()
    """We now have a lock object that will be used to synchronize threads during first
    access to the Singleton."""

    def __call__(cls, *args, **kwargs):
        """
        Return a single instance of the class.

        Possible changes to the value of the ``__init__`` argument do not affect the
        returned instance.
        """
        # Now, imagine that the program has just been launched. Since there's no
        # Singleton instance yet, multiple threads can simultaneously pass the
        # previous conditional and reach this point almost at the same time. The
        # first of them will acquire lock and will proceed further, while the
        # rest will wait here.
        with cls._lock:
            # The first thread to acquire the lock, reaches this conditional,
            # goes inside and creates the Singleton instance. Once it leaves the
            # lock block, a thread that might have been waiting for the lock
            # release may then enter this section. But since the Singleton field
            # is already initialized, the thread won't create a new object.
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class DefaultUnitsClass(metaclass=SingletonMeta):
    """Provides default units for the PyAnsys Geometry singleton design pattern."""

    def __init__(self) -> None:
        """Initialize the ``DefaultUnitsClass`` class."""
        self._length: Unit = UNITS.meter
        self._angle: Unit = UNITS.radian
        self._server_length: Unit = UNITS.meter
        self._server_angle: Unit = UNITS.radian

    @property
    def LENGTH(self) -> Unit:
        """Default length unit for PyAnsys Geometry."""
        return self._length

    @LENGTH.setter
    @check_input_types
    def LENGTH(self, value: Unit) -> None:
        check_pint_unit_compatibility(value, self._length)
        self._length = value

    @property
    def ANGLE(self) -> Unit:
        """Default angle unit for PyAnsys Geometry."""
        return self._angle

    @ANGLE.setter
    @check_input_types
    def ANGLE(self, value: Unit) -> None:
        check_pint_unit_compatibility(value, self._angle)
        self._angle = value

    @property
    def SERVER_LENGTH(self) -> Unit:
        """
        Default length unit for supporting Geometry services for gRPC messages.

        Notes
        -----
        The default units on the server side are not modifiable yet.
        """
        return self._server_length

    @property
    def SERVER_AREA(self) -> Unit:
        """
        Default area unit for supporting Geometry services for gRPC messages.

        Notes
        -----
        The default units on the server side are not modifiable yet.
        """
        return self._server_length * self._server_length

    @property
    def SERVER_VOLUME(self) -> Unit:
        """
        Default volume unit for supporting Geometry services for gRPC messages.

        Notes
        -----
        The default units on the server side are not modifiable yet.
        """
        return self._server_length * self._server_length * self._server_length

    @property
    def SERVER_ANGLE(self) -> Unit:
        """
        Default angle unit for supporting Geometry services for gRPC messages.

        Notes
        -----
        The default units on the server side are not modifiable yet.
        """
        return self._server_angle


DEFAULT_UNITS = DefaultUnitsClass()
"""PyAnsys Geometry default units object."""


class Measurement(PhysicalQuantity):
    """
    Provides the ``PhysicalQuantity`` subclass for holding a measurement.

    Parameters
    ----------
    value : Union[Real, Quantity]
        Value of the measurement.
    unit : ~pint.Unit
        Units for the measurement.
    dimensions : ~pint.Unit
        Units for extracting the dimensions of the measurement.
        If ``~pint.Unit.meter`` is given, the dimension extracted is ``[length]``.
    """

    def __init__(self, value: Union[Real, Quantity], unit: Unit, dimensions: Unit):
        """Initialize the ``Measurement`` class."""
        # Check the input
        if isinstance(value, Quantity):
            # TODO: inform that if Quantity is given, we will ignore provided unit value
            unit = value.units
        else:
            check_is_float_int(value, "value")
            value = Quantity(value, unit)

        # Call the PhysicalQuantity ctor.
        super().__init__(unit, expected_dimensions=dimensions)

        # Store the value
        self._value = self._base_units_magnitude(value)

    @property
    def value(self) -> Quantity:
        """Value of the measurement."""
        return self._get_quantity(self._value)

    @value.setter
    def value(self, value: Quantity) -> None:
        """Set the measurement value."""
        self._value = self._base_units_magnitude(value)

    @check_input_types
    def __eq__(self, other: "Measurement") -> bool:
        """Equals operator for the ``Measurement`` class."""
        return self._value == other._value and self._base_unit == other._base_unit


class Distance(Measurement):
    """
    Provides the ``Measurement`` subclass for holding a distance.

    Parameters
    ----------
    value : Union[Real, Quantity]
        Value of the distance.
    unit : ~pint.Unit, default: DEFAULT_UNITS.LENGTH
        Units for the distance.
    """

    def __init__(self, value: Union[Real, Quantity], unit: Optional[Unit] = None):
        """Initialize the ``Distance`` class."""
        # Delegates in Measurement ctor. forcing expected dimensions.
        unit = unit if unit else DEFAULT_UNITS.LENGTH
        super().__init__(value, unit, DEFAULT_UNITS.LENGTH)


class Angle(Measurement):
    """
    Provides the ``Measurement`` subclass for holding an angle.

    Parameters
    ----------
    value : Union[Real, Quantity]
        Value of the angle.
    unit : ~pint.Unit, default: DEFAULT_UNITS.ANGLE
        Units for the distance.
    """

    def __init__(self, value: Union[Real, Quantity], unit: Optional[Unit] = None):
        """Initialize the ``Angle`` class."""
        # Delegates in Measurement ctor. forcing expected dimensions.
        unit = unit if unit else DEFAULT_UNITS.ANGLE
        super().__init__(value, unit, DEFAULT_UNITS.ANGLE)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides functions for performing common checks."""
from beartype.typing import TYPE_CHECKING, Any, Iterable, Optional, Tuple, Union
import numpy as np
from pint import Unit

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer import Design


def ensure_design_is_active(method):
    """
    Make sure that the design is active before executing a method.

    This function is necessary to be called whenever we do any operation on the design.
    If we are just accessing information of the class, it is not necessary to call this.
    """

    def wrapper(self, *args, **kwargs):
        import ansys.geometry.core as pyansys_geometry

        if pyansys_geometry.DISABLE_MULTIPLE_DESIGN_CHECK:
            # If the user has disabled the check, then we can skip it
            return method(self, *args, **kwargs)

        # Check if the current design is active... otherwise activate it
        def get_design_ref(obj) -> "Design":
            if hasattr(obj, "_modeler"):  # In case of a Design object
                return obj
            elif hasattr(
                obj, "_parent_component"
            ):  # In case of a Body, Component, DesignPoint, Beam
                # Recursive call
                return get_design_ref(obj._parent_component)
            elif hasattr(obj, "_body"):  # In case of a Face, Edge
                # Recursive call
                return get_design_ref(obj._body._parent_component)
            else:
                raise ValueError("Unable to find the design reference.")

        # Get the design reference
        design = get_design_ref(self)

        # Activate the design if it is not active
        if not design.is_active:
            # First, check the backend allows for multiple documents
            if not design._grpc_client.multiple_designs_allowed:
                from ansys.geometry.core.errors import GeometryRuntimeError

                raise GeometryRuntimeError(
                    "The design is not active and multiple designs are "
                    "not allowed with the current backend."
                )
            else:
                design._activate()

        # Finally, call method
        return method(self, *args, **kwargs)

    return wrapper


def check_is_float_int(param: object, param_name: Optional[Union[str, None]] = None) -> None:
    """
    Check if a parameter has a float or integer value.

    Parameters
    ----------
    param : object
        Object instance to check.
    param_name : str, default: None
        Parameter name (if any).

    Raises
    ------
    TypeError
        If the parameter does not have a float or integer value.
    """
    if not isinstance(param, (int, float)):
        raise TypeError(
            f"The parameter should have a float or integer value."
            if param_name is None
            else f"The parameter '{param_name}' should have a float or integer value."
        )


def check_ndarray_is_float_int(
    param: np.ndarray, param_name: Optional[Union[str, None]] = None
) -> None:
    """
    Check if a :class:`numpy.ndarray <numpy.ndarray>` has float or integer values.

    Parameters
    ----------
    param : ~numpy.ndarray
        :class:`numpy.ndarray <numpy.ndarray>` instance to check.
    param_name : str, default: None
        :class:`numpy.ndarray <numpy.ndarray>` instance name (if any).

    Raises
    ------
    TypeError
        If the :class:`numpy.ndarray <numpy.ndarray>` instance does not
        have float or integer values.
    """
    param_data = np.ravel(param)

    if not np.issubdtype(param.dtype, np.number) or not all(
        isinstance(data, (int, float)) for data in param_data.data
    ):
        raise TypeError(
            f"The numpy.ndarray should contain float or integer values."
            if param_name is None
            else f"The numpy.ndarray '{param_name}' should contain float or integer values."
        )


def check_ndarray_is_not_none(
    param: np.ndarray, param_name: Optional[Union[str, None]] = None
) -> None:
    """
    Check if a :class:`numpy.ndarray <numpy.ndarray>` has all ``None`` values.

    Parameters
    ----------
    param : ~numpy.ndarray
        :class:`numpy.ndarray <numpy.ndarray>` instance to check.
    param_name : str, default: None
        :class:`numpy.ndarray <numpy.ndarray>` instance name (if any).

    Raises
    ------
    ValueError
        If the :class:`numpy.ndarray <numpy.ndarray>` instance has a value
        of ``None`` for all parameters.
    """
    param_data = np.ravel(param)
    if all(value is None for value in param_data):
        raise ValueError(
            f"The numpy.ndarray should not have 'None' for all parameter values."
            if param_name is None
            else f"The numpy.ndarray '{param_name}' should not have 'None' for all parameter values."  # noqa: E501
        )


def check_ndarray_is_all_nan(
    param: np.ndarray, param_name: Optional[Union[str, None]] = None
) -> None:
    """
    Check if a :class:`numpy.ndarray <numpy.ndarray>` is all nan-valued.

    Parameters
    ----------
    param : ~numpy.ndarray
        :class:`numpy.ndarray <numpy.ndarray>` instance to check.
    param_name : str or None, default: None
        :class:`numpy.ndarray <numpy.ndarray>` instance name (if any).

    Raises
    ------
    ValueError
        If the :class:`numpy.ndarray <numpy.ndarray>` instance is all nan-valued.
    """
    if np.isnan(param).all():
        raise ValueError(
            f"The numpy.ndarray should not be a nan numpy.ndarray."
            if param_name is None
            else f"The numpy.ndarray '{param_name}' should not be a nan numpy.ndarray."
        )


def check_ndarray_is_non_zero(
    param: np.ndarray, param_name: Optional[Union[str, None]] = None
) -> None:
    """
    Check if a :class:`numpy.ndarray <numpy.ndarray>` is zero-valued.

    Parameters
    ----------
    param : ~numpy.ndarray
        :class:`numpy.ndarray <numpy.ndarray>` instance to check.
    param_name : str, default: None
        :class:`numpy.ndarray <numpy.ndarray>` instance name (if any).

    Raises
    ------
    ValueError
        If the :class:`numpy.ndarray <numpy.ndarray>` instance is zero-valued.
    """
    param_data = np.ravel(param)
    if all(value == 0 for value in param_data):
        raise ValueError(
            f"The numpy.ndarray should not be a numpy.ndarray of zeros."
            if param_name is None
            else f"The numpy.ndarray '{param_name}' should not be a numpy.ndarray of zeros."
        )


def check_pint_unit_compatibility(input: Unit, expected: Unit) -> None:
    """
    Check if input for :class:`pint.Unit` is compatible with the expected input.

    Parameters
    ----------
    input : ~pint.Unit
        :class:`pint.Unit` input.
    expected : ~pint.Unit
        :class:`pint.Unit` expected dimensionality.

    Raises
    ------
    TypeError
        If the input is not compatible with the :class:`pint.Unit` class.
    """
    if not input.is_compatible_with(expected):
        raise TypeError(
            f"The pint.Unit provided as an input should be a {expected.dimensionality} quantity."
        )


def check_type_equivalence(input: object, expected: object) -> None:
    """
    Check if an input object is of the same class as an expected object.

    Parameters
    ----------
    input : object
        Input object.
    expected : object
        Expected object.

    Raises
    ------
    TypeError
        If the objects are not of the same class.
    """
    if not isinstance(input, type(expected)):
        raise TypeError(
            f"Provided type {type(input)} is invalid. Type {type(expected)} is expected."
        )


def check_type(input: object, expected_type: Union[type, Tuple[type, Any]]) -> None:
    """
    Check if an input object is of the same type as expected types.

    Parameters
    ----------
    input : object
        Input object.
    expected_type : Union[type, Tuple[type, ...]]
        One or more types to compare the input object against.

    Raises
    ------
    TypeError
        If the object does not match the one or more expected types.
    """
    if not isinstance(input, expected_type):
        raise TypeError(
            f"Provided type {type(input)} is invalid. Type {expected_type} is expected."
        )


def check_type_all_elements_in_iterable(
    input: Iterable, expected_type: Union[type, Tuple[type, Any]]
) -> None:
    """
    Check if all elements in an iterable are of the same type as expected types.

    Parameters
    ----------
    input : Iterable
        Input iterable.
    expected_type : Union[type, Tuple[type, ...]]
        One or more types to compare the input object against.

    Raises
    ------
    TypeError
        If one of the elements in the iterable does not match the one or more expected types.
    """
    for elem in input:
        check_type(elem, expected_type)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a curve."""
from functools import cached_property

from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import Vector3D
from ansys.geometry.core.typing import Real


class CurveEvaluation:
    """Provides for evaluating a curve."""

    def __init__(self, parameter: Real = None) -> None:
        """Initialize the ``CurveEvaluation`` class."""
        self._parameter = parameter

    def is_set(self) -> bool:
        """Determine if the parameter for the evaluation has been set."""
        return self._parameter is not None

    @property
    def parameter(self) -> Real:
        """Parameter that the evaluation is based upon."""
        raise NotImplementedError("Each evaluation must provide the parameter definition.")

    @cached_property
    def position(self) -> Point3D:
        """Position of the evaluation."""
        raise NotImplementedError("Each evaluation must provide the position definition.")

    @cached_property
    def first_derivative(self) -> Vector3D:
        """First derivative of the evaluation."""
        raise NotImplementedError("Each evaluation must provide the first_derivative definition.")

    @cached_property
    def second_derivative(self) -> Vector3D:
        """Second derivative of the evaluation."""
        raise NotImplementedError("Each evaluation must provide the second_derivative definition.")

    @cached_property
    def curvature(self) -> Real:
        """Curvature of the evaluation."""
        raise NotImplementedError("Each evaluation must provide the curvature definition.")



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing an arc."""

from beartype import beartype as check_input_types
from beartype.typing import Optional
import numpy as np
from pint import Quantity
import pyvista as pv

from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.math.vector import Vector2D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.edge import SketchEdge


class Arc(SketchEdge):
    """
    Provides for modeling an arc.

    Parameters
    ----------
    center : Point2D
        Center point of the arc.
    start : Point2D
        Starting point of the arc.
    end : Point2D
        Ending point of the arc.
    clockwise : bool, default: False
        Whether the arc spans the clockwise angle between the start and end points.
        When ``False`` (default), the arc spans the counter-clockwise angle. When
        ``True``, the arc spands the clockwise angle.
    """

    @check_input_types
    def __init__(
        self,
        center: Point2D,
        start: Point2D,
        end: Point2D,
        clockwise: Optional[bool] = False,
    ):
        """Initialize the arc shape."""
        super().__init__()
        if start == end:
            raise ValueError("Start and end points must be different.")
        if center == start:
            raise ValueError("Center and start points must be different.")
        if center == end:
            raise ValueError("Center and end points must be different.")

        if not np.isclose(
            np.linalg.norm(np.array(start - center)),
            np.linalg.norm(np.array(end - center)),
        ):
            raise ValueError(
                "The start and end points of the arc are not an equidistant from the center point."
            )

        # Store the main three points of the arc
        self._center, self._start, self._end = center, start, end

        # Compute the vectors from the center to the start and end points
        to_start_vector = Vector2D.from_points(start, center)
        to_end_vector = Vector2D.from_points(end, center)

        # Compute the radius of the arc
        self._radius = Quantity(to_start_vector.norm, self._start.base_unit)
        if not self._radius.m > 0:  # pragma: no cover
            raise ValueError("Point configuration does not yield a positive length arc radius.")

        # Compute the angle of the arc (always counter-clockwise at this point)
        self._angle = to_start_vector.get_angle_between(to_end_vector)

        # Check if the clockwise direction is desired... and recompute angle
        self._clockwise = clockwise
        if clockwise:
            self._angle = (2 * np.pi) - self._angle

    @property
    def start(self) -> Point2D:
        """Starting point of the arc line."""
        return self._start

    @property
    def end(self) -> Point2D:
        """Ending point of the arc line."""
        return self._end

    @property
    def length(self) -> Quantity:
        """Length of the arc."""
        return 2 * np.pi * self.radius * (self.angle.m / (2 * np.pi))

    @property
    def radius(self) -> Quantity:
        """Radius of the arc."""
        return self._radius

    @property
    def center(self) -> Point2D:
        """Center point of the arc."""
        return self._center

    @property
    def angle(self) -> Quantity:
        """Angle of the arc."""
        return Quantity(self._angle, UNITS.radian)

    @property
    def is_clockwise(self) -> bool:
        """
        Flag indicating whether the rotation of the angle is clockwise.

        Returns
        -------
        bool
            ``True`` if the sense of rotation is clockwise.
            ``False`` if the sense of rotation is counter-clockwise.
        """
        return self._clockwise

    @property
    def sector_area(self) -> Quantity:
        """Area of the sector of the arc."""
        return self.radius**2 * self.angle.m / 2

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """
        VTK polydata representation for PyVista visualization.

        Notes
        -----
        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        if np.isclose(self.angle, np.pi):
            # TODO : PyVista hack... Maybe worth implementing something in PyVista...
            #        A user should be able to define clockwise/counterclockwise sense of
            #        rotation...
            return self.__arc_pyvista_hack()
        elif self.angle > np.pi:
            pv_negative = True

        else:
            pv_negative = False

        return pv.CircularArc(
            [
                self.start.x.m_as(DEFAULT_UNITS.LENGTH),
                self.start.y.m_as(DEFAULT_UNITS.LENGTH),
                0,
            ],
            [self.end.x.m_as(DEFAULT_UNITS.LENGTH), self.end.y.m_as(DEFAULT_UNITS.LENGTH), 0],
            [
                self.center.x.m_as(DEFAULT_UNITS.LENGTH),
                self.center.y.m_as(DEFAULT_UNITS.LENGTH),
                0,
            ],
            negative=pv_negative,
        )

    @check_input_types
    def __eq__(self, other: "Arc") -> bool:
        """Equals operator for the ``Arc`` class."""
        return bool(
            self.start == other.start
            and self.end == other.end
            and self.center == other.center
            and self.angle == other.angle
        )

    def __ne__(self, other: "Arc") -> bool:
        """Not equals operator for the ``Arc`` class."""
        return not self == other

    def __arc_pyvista_hack(self):
        """
        Hack for close to PI arcs.

        Notes
        -----
        PyVista does not know whether the rotation is
        clockwise or counterclockwise. It only understands the longest and shortest
        angle, which complicates things in the boundary.

        This means that the arc must be divided in two so that it is properly
        defined based on the known sense of rotation.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        # Define the arc mid point
        if not self.is_clockwise:
            rot_matrix = np.array([[0, -1], [1, 0]])  # 90 degs rot matrix
        else:
            rot_matrix = np.array([[0, 1], [-1, 0]])  # -90 degs rot matrix
        center_to_start = Vector2D.from_points(self.center, self.start)
        center_to_mid = rot_matrix @ center_to_start
        mid_point2d = Point2D(
            [
                center_to_mid[0] + self.center.x.to_base_units().m,
                center_to_mid[1] + self.center.y.to_base_units().m,
            ],
            self.center.base_unit,
        )

        # Define auxiliary lists for PyVista containing the start, mid, end, and center points
        mid_point = [
            mid_point2d.x.m_as(DEFAULT_UNITS.LENGTH),
            mid_point2d.y.m_as(DEFAULT_UNITS.LENGTH),
            0,
        ]
        start_point = [
            self.start.x.m_as(DEFAULT_UNITS.LENGTH),
            self.start.y.m_as(DEFAULT_UNITS.LENGTH),
            0,
        ]
        end_point = [
            self.end.x.m_as(DEFAULT_UNITS.LENGTH),
            self.end.y.m_as(DEFAULT_UNITS.LENGTH),
            0,
        ]
        center_point = [
            self.center.x.m_as(DEFAULT_UNITS.LENGTH),
            self.center.y.m_as(DEFAULT_UNITS.LENGTH),
            0,
        ]

        # Compute the two independent arcs
        arc_sub1 = pv.CircularArc(
            start_point,
            mid_point,
            center_point,
        )
        arc_sub2 = pv.CircularArc(
            mid_point,
            end_point,
            center_point,
        )

        return arc_sub1 + arc_sub2

    @classmethod
    def from_three_points(cls, start: Point2D, inter: Point2D, end: Point2D):
        """
        Create an arc from three given points.

        Parameters
        ----------
        start : Point2D
            Starting point of the arc.
        inter : Point2D
            Intermediate point (location) of the arc.
        end : Point2D
            Ending point of the arc.

        Returns
        -------
        Arc
            Arc generated from the three points.
        """
        # Unpack the points into its coordinates (in DEFAULT_UNITS.LENGTH)
        x_s, y_s = start.tolist()
        x_i, y_i = inter.tolist()
        x_e, y_e = end.tolist()

        # Solve the circle equation to find out its center
        # (Xc - X)**2 + (Yc - Y)**2 = r**2
        #
        # Since r is the radius, and its constant, you can just simply solve
        # the system of equations formed by the three points, for (x_c, y_c)
        #
        # 2*(x_s - x_i)*x_c + 2*(y_s-y_i)*y_c + (x_i**2 + y_i**2 - x_s**2 - y_s**2) = 0
        # 2*(x_s - x_e)*x_c + 2*(y_s-y_e)*y_c + (x_e**2 + y_e**2 - x_s**2 - y_s**2) = 0
        #
        # Or...
        #
        # k11 * x_c + k12 * y_c = k1
        # k21 * x_c + k22 * y_c = k2
        #
        # Where...
        #
        # k11 = 2*(x_s - x_i)
        # k12 = 2*(y_s - y_i)
        # k21 = 2*(x_s - x_e)
        # k22 = 2*(y_s - y_e)
        #
        # k1 = (x_s**2 + y_s**2) - (x_i**2 + y_i**2)
        # k2 = (x_s**2 + y_s**2) - (x_e**2 + y_e**2)
        #
        # This is easily done with numpy!
        #
        k11 = 2 * (x_s - x_i)
        k12 = 2 * (y_s - y_i)
        k21 = 2 * (x_s - x_e)
        k22 = 2 * (y_s - y_e)

        k1 = (x_s**2 + y_s**2) - (x_i**2 + y_i**2)
        k2 = (x_s**2 + y_s**2) - (x_e**2 + y_e**2)

        x_c, y_c = np.linalg.solve([[k11, k12], [k21, k22]], [k1, k2])
        center = Point2D([x_c, y_c], unit=DEFAULT_UNITS.LENGTH)

        # Now, you should try to figure out if the rotation has to be clockwise or
        # counter-clockwise...
        center_start = Vector2D([x_s - x_c, y_s - y_c])
        center_inter = Vector2D([x_i - x_c, y_i - y_c])
        center_end = Vector2D([x_e - x_c, y_e - y_c])
        angle_s_i = Vector2D.get_angle_between(center_start, center_inter)
        angle_s_e = Vector2D.get_angle_between(center_start, center_end)

        is_clockwise = False if angle_s_i < angle_s_e else True

        # Finally... you can create the arc
        return Arc(center=center, start=start, end=end, clockwise=is_clockwise)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides plotting for various PyAnsys Geometry objects."""
import re

from beartype.typing import Any, Dict, List, Optional, Union
import numpy as np
import pyvista as pv
from pyvista.plotting.plotter import Plotter as PyVistaPlotter
from pyvista.plotting.tools import create_axes_marker

from ansys.geometry.core import DOCUMENTATION_BUILD
from ansys.geometry.core.designer.body import Body, MasterBody
from ansys.geometry.core.designer.component import Component
from ansys.geometry.core.designer.design import Design
from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.logger import LOG as logger
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.plotting.plotting_types import EdgePlot, GeomObjectPlot
from ansys.geometry.core.sketch.sketch import Sketch

DEFAULT_COLOR = "#D6F7D1"
"""Default color we use for the plotter actors."""

PICKED_COLOR = "#BB6EEE"
"""Color to use for the actors that are currently picked."""

EDGE_COLOR = "#000000"
"""Default color to use for the edges."""

PICKED_EDGE_COLOR = "#9C9C9C"
"""Color to use for the edges that are currently picked."""


class Plotter:
    """
    Provides for plotting sketches and bodies.

    Parameters
    ----------
    scene : ~pyvista.Plotter, default: None
        ``Scene`` instance for rendering the objects.
    color_opts : dict, default: None
        Dictionary containing the background and top colors.
    num_points : int, default: 100
        Number of points to use to render the shapes.
    enable_widgets: bool, default: True
        Whether to enable widget buttons in the plotter window.
        Widget buttons must be disabled when using
        `trame <https://kitware.github.io/trame/index.html>`_
        for visualization.
    """

    def __init__(
        self,
        scene: Optional[pv.Plotter] = None,
        color_opts: Optional[Dict] = None,
        num_points: int = 100,
        enable_widgets: bool = True,
    ) -> None:
        """Initialize the plotter."""
        # Generate custom scene if ``None`` is provided
        if scene is None:
            scene = pv.Plotter()

        # If required, use a white background with no gradient
        if not color_opts:
            color_opts = dict(color="white")

        # Create the scene
        self._scene = scene
        # Scene: assign the background
        self._scene.set_background(**color_opts)
        view_box = self._scene.add_axes(line_width=5, color="black")

        # Save the desired number of points
        self._num_points = num_points

        # geometry objects to actors mapping
        self._geom_object_actors_map = {}
        self._enable_widgets = enable_widgets

    @property
    def scene(self) -> PyVistaPlotter:
        """
        Rendered scene object.

        Returns
        -------
        ~pyvista.Plotter
            Rendered scene object.
        """
        return self._scene

    def view_xy(self) -> None:
        """View the scene from the XY plane."""
        self.scene.view_xy()

    def view_xz(self) -> None:
        """View the scene from the XZ plane."""
        self.scene.view_xz()

    def view_yx(self) -> None:
        """View the scene from the YX plane."""
        self.scene.view_yx()

    def view_yz(self) -> None:
        """View the scene from the YZ plane."""
        self.scene.view_yz()

    def view_zx(self) -> None:
        """View the scene from the ZX plane."""
        self.scene.view_zx()

    def view_zy(self) -> None:
        """View the scene from the ZY plane."""
        self.scene.view_zy()

    def plot_frame(self, frame: Frame, plotting_options: Optional[Dict] = None) -> None:
        """
        Plot a frame in the scene.

        Parameters
        ----------
        frame : Frame
            Frame to render in the scene.
        plotting_options : dict, default: None
            Dictionary containing parameters accepted by the
            :func:`pyvista.create_axes_marker` class for customizing
            the frame rendering in the scene.
        """
        # Use default plotting options if required
        if plotting_options is None:
            plotting_options = dict(labels_off=True, cone_radius=0.2)

        # Create the axes actor
        axes = create_axes_marker(**plotting_options)

        # Transpose the matrix for fixing rotation sense in VTK
        arr = np.vstack((frame.global_to_local_rotation, frame.origin)).T
        arr = np.vstack((arr, [0, 0, 0, 1]))

        # Apply matrix transformation to the actor
        axes.SetUserMatrix(pv.vtkmatrix_from_array(arr))

        # Render the actor in the scene
        self.scene.add_actor(axes)

    def plot_plane(
        self,
        plane: Plane,
        plane_options: Optional[Dict] = None,
        plotting_options: Optional[Dict] = None,
    ) -> None:
        """
        Plot a plane in the scene.

        Parameters
        ----------
        plane : Plane
            Plane to render in the scene.
        plane_options : dict, default: None
            Dictionary containing parameters accepted by the
            :func:`pyvista.Plane  <pyvista.Plane>` function for customizing the mesh
            representing the plane.
        plotting_options : dict, default: None
            Dictionary containing parameters accepted by the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method for
            customizing the mesh rendering of the plane.
        """
        # Impose default plane options if none provided
        if plane_options is None:
            plane_options = dict(i_size=10, j_size=10)

        plane_mesh = pv.Plane(
            center=plane.origin.tolist(), direction=plane.direction_z.tolist(), **plane_options
        )

        # Impose default plotting options if none provided
        if plotting_options is None:
            plotting_options = dict(color="blue", opacity=0.1)

        self.scene.add_mesh(plane_mesh, **plotting_options)

    def plot_sketch(
        self,
        sketch: Sketch,
        show_plane: bool = False,
        show_frame: bool = False,
        **plotting_options: Optional[Dict],
    ) -> None:
        """
        Plot a sketch in the scene.

        Parameters
        ----------
        sketch : Sketch
            Sketch to render in the scene.
        show_plane : bool, default: False
            Whether to render the sketch plane in the scene.
        show_frame : bool, default: False
            Whether to show the frame in the scene.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Show the sketch plane if required
        if show_plane:
            self.plot_plane(sketch._plane)

        # Show the sketch plane if required
        if show_frame:
            self.plot_frame(sketch._plane)

        if "clipping_plane" in plotting_options:
            logger.warning("Clipping is not available in Sketch objects.")
            plotting_options.pop("clipping_plane")

        self.add_sketch_polydata(sketch.sketch_polydata_faces(), opacity=0.7, **plotting_options)
        self.add_sketch_polydata(sketch.sketch_polydata_edges(), **plotting_options)

    def add_body_edges(self, body_plot: GeomObjectPlot, **plotting_options: Optional[dict]) -> None:
        """
        Add the outer edges of a body to the plot.

        This method has the side effect of adding the edges to the GeomObject that
        you pass through the parameters.

        Parameters
        ----------
        body : GeomObjectPlot
            Body of which to add the edges.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        edge_plot_list = []
        for edge in body_plot.object.edges:
            line = pv.Line(edge.start_point, edge.end_point)
            edge_actor = self.scene.add_mesh(
                line, line_width=10, color=EDGE_COLOR, **plotting_options
            )
            edge_actor.SetVisibility(False)
            edge_plot = EdgePlot(edge_actor, edge, body_plot)
            edge_plot_list.append(edge_plot)
        body_plot.edges = edge_plot_list

    def add_body(
        self, body: Body, merge: Optional[bool] = False, **plotting_options: Optional[Dict]
    ) -> None:
        """
        Add a body to the scene.

        Parameters
        ----------
        body : Body
            Body to add.
        merge : bool, default: False
            Whether to merge the body into a single mesh. When ``True``, the
            individual faces of the tessellation are merged. This
            preserves the number of triangles and only merges the topology.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments,
            see the :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Use the default PyAnsys Geometry add_mesh arguments
        self.__set_add_mesh_defaults(plotting_options)
        dataset = body.tessellate(merge=merge)
        if "clipping_plane" in plotting_options:
            dataset = self.clip(dataset, plotting_options.get("clipping_plane"))
            plotting_options.pop("clipping_plane", None)
        if isinstance(dataset, pv.MultiBlock):
            actor, _ = self.scene.add_composite(dataset, **plotting_options)
        else:
            actor = self.scene.add_mesh(dataset, **plotting_options)

        body_plot = GeomObjectPlot(actor=actor, object=body, add_body_edges=True)
        self.add_body_edges(body_plot)
        self._geom_object_actors_map[actor] = body_plot

    def add_component(
        self,
        component: Component,
        merge_component: bool = False,
        merge_bodies: bool = False,
        **plotting_options,
    ) -> str:
        """
        Add a component to the scene.

        Parameters
        ----------
        component : Component
            Component to add.
        merge_component : bool, default: False
            Whether to merge the component into a single dataset. When
            ``True``, all the individual bodies are effectively combined
            into a single dataset without any hierarchy.
        merge_bodies : bool, default: False
            Whether to merge each body into a single dataset. When ``True``,
            all the faces of each individual body are effectively combined
            into a single dataset without separating faces.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.

        Returns
        -------
        str
            Name of the added PyVista actor.
        """
        # Use the default PyAnsys Geometry add_mesh arguments
        self.__set_add_mesh_defaults(plotting_options)
        dataset = component.tessellate(merge_component=merge_component, merge_bodies=merge_bodies)

        if "clipping_plane" in plotting_options:
            dataset = self.clip(dataset, plotting_options["clipping_plane"])
            plotting_options.pop("clipping_plane", None)

        if isinstance(dataset, pv.MultiBlock):
            actor, _ = self.scene.add_composite(dataset, **plotting_options)
        else:
            actor = self.scene.add_mesh(dataset, **plotting_options)

        return actor.name

    def add_sketch_polydata(self, polydata_entries: List[pv.PolyData], **plotting_options) -> None:
        """
        Add sketches to the scene from PyVista polydata.

        Parameters
        ----------
        polydata : pyvista.PolyData
            Polydata to add.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Use the default PyAnsys Geometry add_mesh arguments
        mb = pv.MultiBlock()
        for polydata in polydata_entries:
            mb.append(polydata)

        if "clipping_plane" in plotting_options:
            mb = self.clip(mb, plane=plotting_options["clipping_plane"])
            plotting_options.pop("clipping_plane", None)

        self.scene.add_mesh(mb, color=EDGE_COLOR, **plotting_options)

    def clip(
        self, mesh: Union[pv.PolyData, pv.MultiBlock], plane: Plane = None
    ) -> Union[pv.PolyData, pv.MultiBlock]:
        """
        Clip the passed mesh with a plane.

        Parameters
        ----------
        mesh : Union[pv.PolyData, pv.MultiBlock]
            Mesh you want to clip.
        normal : str, optional
            Plane you want to use for clipping, by default "x".
            Available options: ["x", "-x", "y", "-y", "z", "-z"]
        origin : tuple, optional
            Origin point of the plane, by default None

        Returns
        -------
        Union[pv.PolyData,pv.MultiBlock]
            The clipped mesh.
        """
        return mesh.clip(normal=plane.normal, origin=plane.origin)

    def add_design_point(self, design_point: DesignPoint, **plotting_options) -> None:
        """
        Add a DesignPoint object to the plotter.

        Parameters
        ----------
        design_point : DesignPoint
            DesignPoint to add.
        """
        # get the actor for the DesignPoint
        actor = self.scene.add_mesh(design_point._to_polydata(), **plotting_options)

        # save the actor to the object/actor map
        design_point = GeomObjectPlot(actor=actor, object=design_point, add_body_edges=False)

        self._geom_object_actors_map[actor] = design_point

    def add(
        self,
        object: Any,
        merge_bodies: bool = False,
        merge_components: bool = False,
        filter: str = None,
        **plotting_options,
    ) -> None:
        """
        Add any type of object to the scene.

        These types of objects are supported: ``Body``, ``Component``, ``List[pv.PolyData]``,
        ``pv.MultiBlock``, and ``Sketch``.

        Parameters
        ----------
        plotting_list : List[Any]
            List of objects that you want to plot.
        merge_bodies : bool, default: False
            Whether to merge each body into a single dataset. When ``True``,
            all the faces of each individual body are effectively combined
            into a single dataset without separating faces.
        merge_component : bool, default: False
            Whether to merge the component into a single dataset. When
            ``True``, all the individual bodies are effectively combined
            into a single dataset without any hierarchy.
        filter : str, default: None
            Regular expression with the desired name or names you want to include in the plotter.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        logger.debug(f"Adding object type {type(object)} to the PyVista plotter")
        if filter:
            if hasattr(object, "name"):
                if not re.search(filter, object.name):
                    logger.debug(f"Name {object.name} not found in regex {filter}.")
                    return self._geom_object_actors_map

        # Check what kind of object we are dealing with
        if isinstance(object, List) and isinstance(object[0], pv.PolyData):
            self.add_sketch_polydata(object, **plotting_options)
        elif isinstance(object, pv.PolyData):
            if "clipping_plane" in plotting_options:
                object = self.clip(object, plotting_options["clipping_plane"])
                plotting_options.pop("clipping_plane", None)
            self.scene.add_mesh(object, **plotting_options)
        elif isinstance(object, pv.MultiBlock):
            if "clipping_plane" in plotting_options:
                object = self.clip(object, plotting_options["clipping_plane"])
                plotting_options.pop("clipping_plane", None)
            self.scene.add_composite(object, **plotting_options)
        elif isinstance(object, Sketch):
            self.plot_sketch(object, **plotting_options)
        elif isinstance(object, Body) or isinstance(object, MasterBody):
            self.add_body(object, merge_bodies, **plotting_options)
        elif isinstance(object, Design) or isinstance(object, Component):
            self.add_component(object, merge_components, merge_bodies, **plotting_options)
        elif isinstance(object, DesignPoint):
            self.add_design_point(object, **plotting_options)
        else:
            logger.warning(f"Object type {type(object)} can not be plotted.")

    def add_list(
        self,
        plotting_list: List[Any],
        merge_bodies: bool = False,
        merge_components: bool = False,
        filter: str = None,
        **plotting_options,
    ) -> None:
        """
        Add a list of any type of object to the scene.

        These types of objects are supported: ``Body``, ``Component``, ``List[pv.PolyData]``,
        ``pv.MultiBlock``, and ``Sketch``.

        Parameters
        ----------
        plotting_list : List[Any]
            List of objects you want to plot.
        merge_component : bool, default: False
            Whether to merge the component into a single dataset. When
            ``True``, all the individual bodies are effectively combined
            into a single dataset without any hierarchy.
        merge_bodies : bool, default: False
            Whether to merge each body into a single dataset. When ``True``,
            all the faces of each individual body are effectively combined
            into a single dataset without separating faces.
        filter : str, default: None
            Regular expression with the desired name or names you want to include in the plotter.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        for object in plotting_list:
            _ = self.add(object, merge_bodies, merge_components, filter, **plotting_options)

    def show(
        self,
        show_axes_at_origin: bool = True,
        show_plane: bool = True,
        jupyter_backend: Optional[str] = None,
        **kwargs: Optional[Dict],
    ) -> None:
        """
        Show the rendered scene on the screen.

        Parameters
        ----------
        jupyter_backend : str, default: None
            PyVista Jupyter backend.
        **kwargs : dict, default: None
            Plotting keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.show <pyvista.Plotter.show>` method.

        Notes
        -----
        For more information on supported Jupyter backends, see
        `Jupyter Notebook Plotting <https://docs.pyvista.org/user-guide/jupyter/index.html>`_
        in the PyVista documentation.
        """
        # computue the scaling
        bounds = self.scene.renderer.bounds
        x_length, y_length = bounds[1] - bounds[0], bounds[3] - bounds[2]
        sfac = max(x_length, y_length)

        # # Show origin axes without labels
        # if show_axes_at_origin:
        #     axes = create_axes_marker(labels_off=True)
        #     self.scene.add_actor(axes)

        # Create the fundamental XY plane
        if show_plane:
            # self.scene.bounds
            plane = pv.Plane(i_size=sfac * 1.3, j_size=sfac * 1.3)
            self.scene.add_mesh(plane, color="white", show_edges=True, opacity=0.1)

        # Conditionally set the Jupyter backend as not all users will be within
        # a notebook environment to avoid a pyvista warning
        if self.scene.notebook and jupyter_backend is None:
            jupyter_backend = "trame"

        # Override jupyter backend in case we are building docs
        if DOCUMENTATION_BUILD:
            jupyter_backend = "static"

        # Enabling anti-aliasing by default on scene
        self.scene.enable_anti_aliasing("ssaa")

        self.scene.show(jupyter_backend=jupyter_backend, **kwargs)

    def __set_add_mesh_defaults(self, plotting_options: Optional[Dict]) -> None:
        # If the following keys do not exist, set the default values
        #
        # This method should only be applied in 3D objects: bodies, components
        plotting_options.setdefault("smooth_shading", True)
        plotting_options.setdefault("color", DEFAULT_COLOR)

    @property
    def geom_object_actors_map(self) -> Dict[pv.Actor, GeomObjectPlot]:
        """Mapping between the ~pyvista.Actor and the PyAnsys Geometry objects."""
        return self._geom_object_actors_map



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for interacting with the Geometry service."""
import logging
import os
from pathlib import Path

from ansys.api.dbu.v0.dbuapplication_pb2 import RunScriptFileRequest
from ansys.api.dbu.v0.dbuapplication_pb2_grpc import DbuApplicationStub
from ansys.api.dbu.v0.designs_pb2 import OpenRequest
from ansys.api.dbu.v0.designs_pb2_grpc import DesignsStub
from ansys.api.geometry.v0.commands_pb2 import UploadFileRequest
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from beartype.typing import TYPE_CHECKING, Dict, Optional, Tuple, Union
from grpc import Channel

from ansys.geometry.core.connection.backend import BackendType
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.defaults import DEFAULT_HOST, DEFAULT_PORT
from ansys.geometry.core.errors import GeometryRuntimeError, protect_grpc
from ansys.geometry.core.logger import LOG as logger
from ansys.geometry.core.misc.checks import check_type
from ansys.geometry.core.misc.options import ImportOptions
from ansys.geometry.core.tools.repair_tools import RepairTools
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from ansys.platform.instancemanagement import Instance

    from ansys.geometry.core.connection.docker_instance import LocalDockerInstance
    from ansys.geometry.core.connection.product_instance import ProductInstance
    from ansys.geometry.core.designer.design import Design


class Modeler:
    """
    Provides for interacting with an open session of the Geometry service.

    Parameters
    ----------
    host : str,  default: DEFAULT_HOST
        Host where the server is running.
    port : Union[str, int], default: DEFAULT_PORT
        Port number where the server is running.
    channel : ~grpc.Channel, default: None
        gRPC channel for server communication.
    remote_instance : ansys.platform.instancemanagement.Instance, default: None
        Corresponding remote instance when the Geometry service
        is launched using `PyPIM <https://github.com/ansys/pypim>`_. This instance
        is deleted when the :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close>`
        method is called.
    docker_instance : LocalDockerInstance, default: None
        Corresponding local Docker instance when the Geometry service is launched using the
        :func:`launch_docker_modeler<ansys.geometry.core.connection.launcher.launch_docker_modeler>`
        method. This instance is deleted when the
        :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close>`
        method is called.
    product_instance : ProductInstance, default: None
        Corresponding local product instance when the product (Discovery or SpaceClaim)
        is launched through the ``launch_modeler_with_geometry_service()``,
        ``launch_modeler_with_discovery()`` or the ``launch_modeler_with_spaceclaim()``
        interface. This instance will be deleted
        when the :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close >`
        method is called.
    timeout : Real, default: 120
        Time in seconds for trying to achieve the connection.
    logging_level : int, default: INFO
        Logging level to apply to the client.
    logging_file : str, Path, default: None
        File to output the log to, if requested.
    """

    def __init__(
        self,
        host: str = DEFAULT_HOST,
        port: Union[str, int] = DEFAULT_PORT,
        channel: Optional[Channel] = None,
        remote_instance: Optional["Instance"] = None,
        docker_instance: Optional["LocalDockerInstance"] = None,
        product_instance: Optional["ProductInstance"] = None,
        timeout: Optional[Real] = 120,
        logging_level: Optional[int] = logging.INFO,
        logging_file: Optional[Union[Path, str]] = None,
        backend_type: Optional[BackendType] = None,
    ):
        """Initialize the ``Modeler`` class."""
        self._client = GrpcClient(
            host=host,
            port=port,
            channel=channel,
            remote_instance=remote_instance,
            docker_instance=docker_instance,
            product_instance=product_instance,
            timeout=timeout,
            logging_level=logging_level,
            logging_file=logging_file,
            backend_type=backend_type,
        )

        # Initialize the RepairTools - Not available on Linux
        # TODO: delete "if" when Linux service is able to use repair tools
        if self.client.backend_type == BackendType.LINUX_SERVICE:
            self._repair_tools = None
            logger.warning("Linux backend does not support repair tools.")
        else:
            self._repair_tools = RepairTools(self._client)

        # Maintaining references to all designs within the modeler workspace
        self._designs: Dict[str, "Design"] = {}

        # Check if the backend allows for multiple designs and throw warning if needed
        if not self.client.multiple_designs_allowed:
            logger.warning(
                "Linux and Ansys Discovery backends do not support multiple "
                "designs open in the same session. Only the last design created "
                "will be available to perform modeling operations."
            )

    @property
    def client(self) -> GrpcClient:
        """``Modeler`` instance client."""
        return self._client

    def create_design(self, name: str) -> "Design":
        """
        Initialize a new design with the connected client.

        Parameters
        ----------
        name : str
            Name for the new design.

        Returns
        -------
        Design
            Design object created on the server.
        """
        from ansys.geometry.core.designer.design import Design

        check_type(name, str)
        design = Design(name, self)
        self._designs[design.design_id] = design
        if len(self._designs) > 1:
            logger.warning(
                "Some backends only support one design. "
                + "Previous designs may be deleted (on the service) when creating a new one."
            )
        return self._designs[design.design_id]

    def get_active_design(self, sync_with_backend: bool = True) -> "Design":
        """
        Get the active design on the modeler object.

        Parameters
        ----------
        sync_with_backend : bool, default: True
            Whether to sync the active design with the remote service. If set to False,
            the active design may be out-of-sync with the remote service. This is useful
            when the active design is known to be up-to-date.

        Returns
        -------
        Design
            Design object already existing on the modeler.
        """
        for _, design in self._designs.items():
            if design._is_active:

                # Check if sync_with_backend is requested
                if sync_with_backend:
                    design._update_design_inplace()

                # Return the active design
                return design

        return None

    def read_existing_design(self) -> "Design":
        """
        Read the existing design on the service with the connected client.

        Returns
        -------
        Design
            Design object already existing on the server.
        """
        from ansys.geometry.core.designer.design import Design

        design = Design("", self, read_existing_design=True)
        self._designs[design.design_id] = design
        if len(self._designs) > 1:
            logger.warning(
                "Some backends only support one design. "
                + "Previous designs may be deleted (on the service) when reading a new one."
            )
        return self._designs[design.design_id]

    def close(self) -> None:
        """``Modeler`` method for easily accessing the client's close method."""
        return self.client.close()

    def _upload_file(
        self,
        file_path: str,
        open_file: bool = False,
        import_options: ImportOptions = ImportOptions(),
    ) -> str:
        """
        Upload a file from the client to the server.

        Notes
        -----
        This method creates a file on the server that has the same name and extension
        as the file on the client.

        Parameters
        ----------
        file_path : str
            Path of the file to upload. The extension of the file must be included.
        open_file : bool, default: False
            Whether to open the file in the Geometry service.
        import_options : ImportOptions
            Import options that toggle certain features when opening a file.

        Returns
        -------
        file_path : str
            Full path of the file uploaded to the server.
        """
        import os

        if not os.path.exists(file_path):
            raise ValueError(f"Could not find file: {file_path}")
        if os.path.isdir(file_path):
            raise ValueError("File path must lead to a file, not a directory.")

        file_name = os.path.split(file_path)[1]

        with open(file_path, "rb") as file:
            data = file.read()

        c_stub = CommandsStub(self._client.channel)

        response = c_stub.UploadFile(
            UploadFileRequest(
                data=data,
                file_name=file_name,
                open=open_file,
                import_options=import_options.to_dict(),
            )
        )
        return response.file_path

    @protect_grpc
    def open_file(
        self,
        file_path: str,
        upload_to_server: bool = True,
        import_options: ImportOptions = ImportOptions(),
    ) -> "Design":
        """
        Open a file.

        This method imports a design into the service. On Windows, ``.scdocx``
        and HOOPS Exchange formats are supported. On Linux, only the ``.scdocx``
        format is supported.

        If the file is a shattered assembly with external references, the whole containing folder
        will need to be uploaded. Ensure proper folder structure in order to prevent the uploading
        of unnecessary files.

        Parameters
        ----------
        file_path : str
            Path of the file to open. The extension of the file must be included.
        upload_to_server : bool
            True if the service is running on a remote machine. If service is running on the local
            machine, set to False, as there is no reason to upload the file.
        import_options : ImportOptions
            Import options that toggle certain features when opening a file.

        Returns
        -------
        Design
            Newly imported design.
        """
        # Format-specific logic - upload the whole containing folder for assemblies
        if upload_to_server:
            if any(
                ext in str(file_path) for ext in [".CATProduct", ".asm", ".solution", ".sldasm"]
            ):
                dir = os.path.dirname(file_path)
                files = os.listdir(dir)
                for file in files:
                    full_path = os.path.join(dir, file)
                    if full_path != file_path:
                        self._upload_file(full_path)
            self._upload_file(file_path, True, import_options)
        else:
            DesignsStub(self._client.channel).Open(
                OpenRequest(filepath=file_path, import_options=import_options.to_dict())
            )

        return self.read_existing_design()

    def __repr__(self) -> str:
        """Represent the modeler as a string."""
        lines = []
        lines.append(f"Ansys Geometry Modeler ({hex(id(self))})")
        lines.append("")
        lines.append(str(self._client))
        return "\n".join(lines)

    @protect_grpc
    def run_discovery_script_file(
        self, file_path: str, script_args: Optional[Dict[str, str]] = None, import_design=False
    ) -> Tuple[Dict[str, str], Optional["Design"]]:
        """
        Run a Discovery script file.

        .. note::

            If arguments are passed to the script, they must be in the form of a dictionary.
            On the server side, the script will receive the arguments as a dictionary of strings,
            under the variable name ``argsDict``. For example, if the script is called with the
            arguments ``run_discovery_script_file(..., script_args = {"length": "20"}, ...)``,
            the script will receive the dictionary ``argsDict`` with the key-value pair
            ``{"length": "20"}``.

        .. note::

            If an output is expected from the script, it will be returned as a dictionary of
            strings. The keys and values of the dictionary are the variables and their values
            that the script returns. However, it is necessary that the script creates a
            dictionary called ``result`` with the variables and their values that are expected
            to be returned. For example, if the script is expected to return the number of bodies
            in the design, the script should create a dictionary called ``result`` with the
            key-value pair ``{"numBodies": numBodies}``, where ``numBodies`` is the number of
            bodies in the design.

        The implied API version of the script should match the API version of the running
        Geometry Service. DMS API versions 23.2.1 and later are supported. DMS is a
        Windows-based modeling service that has been containerized to ease distribution,
        execution, and remotability operations.

        Parameters
        ----------
        file_path : str
            Path of the file. The extension of the file must be included.
        script_args : Optional[Dict[str, str]], optional.
            Arguments to pass to the script. By default, ``None``.
        import_design : bool, optional.
            Whether to refresh the current design from the service. When the script
            is expected to modify the existing design, set this to ``True`` to retrieve
            up-to-date design data. When this is set to ``False`` (default) and the
            script modifies the current design, the design may be out-of-sync. By default,
            ``False``.

        Returns
        -------
        dict[str, str]
            Values returned from the script.
        Design, optional
            Up-to-date current design. This is only returned if ``import_design=True``.

        Raises
        ------
        GeometryRuntimeError
            If the Discovery script fails to run. Otherwise, assume that the script
            ran successfully.
        """
        serv_path = self._upload_file(file_path)
        ga_stub = DbuApplicationStub(self._client.channel)
        request = RunScriptFileRequest(
            script_path=serv_path,
            script_args=script_args,
        )

        self.client.log.debug(f"Running Discovery script file at {file_path}...")
        response = ga_stub.RunScriptFile(request)

        if not response.success:
            raise GeometryRuntimeError(response.message)

        self.client.log.debug(f"Script result message: {response.message}")

        if import_design:
            return (response.values, self.read_existing_design())
        else:
            return response.values

    @property
    def repair_tools(self) -> RepairTools:
        """Access to repair tools."""
        return self._repair_tools



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for creating and managing design points."""

from beartype.typing import TYPE_CHECKING, Union

from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.misc.checks import check_type
from ansys.geometry.core.misc.units import UNITS

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv

    from ansys.geometry.core.designer.component import Component


class DesignPoint:
    """
    Provides for creating design points in components.

    Parameters
    ----------
    id : str
        Server-defined ID for the design points.
    name : str
        User-defined label for the design points.
    points : Point3D
        3D point constituting the design points.
    parent_component : Component
        Parent component to place the new design point under within the design assembly.
    """

    def __init__(self, id: str, name: str, point: Point3D, parent_component: "Component"):
        """Initialize the ``DesignPoints`` class."""
        from ansys.geometry.core.designer.component import Component

        check_type(id, str)
        check_type(name, str)
        check_type(point, Point3D)
        check_type(parent_component, Component)

        self._id = id
        self._name = name
        self._value = point
        self._parent_component = parent_component

    @property
    def id(self) -> str:
        """ID of the design point."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the design point."""
        return self._name

    @property
    def value(self) -> Point3D:
        """Value of the design point."""
        return self._value

    @property
    def parent_component(self) -> Union["Component", None]:
        """Component node that the design point is under."""
        return self._parent_component

    def __repr__(self) -> str:
        """Represent the design points as a string."""
        lines = [f"ansys.geometry.core.designer.DesignPoints {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Design Point         : {self.value}")
        return "\n".join(lines)

    def _to_polydata(self) -> "pv.PolyData":
        """Get polydata from DesignPoint object."""
        import pyvista as pv

        # get units to plot proportionally
        # 0.3 is the size for the sphere representation
        # determined empirically for proper representation
        unit = 0.3 * self.value.unit
        return pv.Sphere(center=self.value.flat, radius=unit.to(UNITS.m).magnitude)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing an edge."""

from beartype.typing import TYPE_CHECKING
from pint import Quantity
import pyvista as pv

from ansys.geometry.core.math.point import Point2D

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.math.plane import Plane


class SketchEdge:
    """Provides for modeling edges forming sketched shapes."""

    @property
    def start(self) -> Point2D:
        """Starting point of the edge."""
        raise NotImplementedError("Each edge must provide the start point definition.")

    @property
    def end(self) -> Point2D:
        """Ending point of the edge."""
        raise NotImplementedError("Each edge must provide the end point definition.")

    @property
    def length(self) -> Quantity:
        """Length of the edge."""
        raise NotImplementedError("Each edge must provide the length definition.")

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """
        VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        raise NotImplementedError("Each edge must provide the polydata definition.")

    def plane_change(self, plane: "Plane") -> None:
        """
        Redefine the plane containing ``SketchEdge`` objects.

        Notes
        -----
        This implies that their 3D definition might suffer changes. By default, this
        metho does nothing. It is required to be implemented in child ``SketchEdge``
        classes.

        Parameters
        ----------
        plane : Plane
            Desired new plane that is to contain the sketched edge.
        """
        pass
